@@ -57,6 +57,7 @@ void network_connect_failed(struct network *network, bool in_handshake);
 bool network_bss_add(struct network *network, struct scan_bss *bss);
 bool network_bss_list_isempty(struct network *network);
 void network_bss_list_clear(struct network *network);
+struct scan_bss *network_bss_list_pop(struct network *network);
 struct scan_bss *network_bss_find_by_addr(struct network *network,
 							const uint8_t *addr);
 struct scan_bss *network_bss_select(struct network *network,
@@ -69,8 +70,9 @@ void network_remove(struct network *network, int reason);
 int network_rank_compare(const void *a, const void *b, void *user);
 void network_rank_update(struct network *network, bool connected);
 
-void network_connect_new_hidden_network(struct network *network,
-					struct l_dbus_message **message);
+struct l_dbus_message *network_connect_new_hidden_network(
+						struct network *network,
+						struct l_dbus_message *message);
 
 void network_blacklist_add(struct network *network, struct scan_bss *bss);
 
